r"""
REFERENCES:

- [CD] -- M. Cheng, J. Duncan - The Largest Mathieu Group and (Mock) Automorphic Forms
- [Ra] -- M. Raum - Powers of M_24-twisted Siegel Products Expansions are Modular.

Recall that weak Jacobi forms of weight 0 and index 1
are storted as pairs (tc_1, tc_2), where `tc_1` is a constant
and `tc_2` is a modular form of weight 2.
In order to unify the implementation, `tc_1` is considered as a modular
form of weight 0.

Initial data coming from [CD]:

- cheng_Ng contains `N_g`.
- cheng_tdT_level contains the level of `{\td T}_g`.
- cheng_chi contains `\chi_g`.
- cheng_cycle_powers describes the powers of conjugacy classes;
  see the definition for details.
- cheng_Zg(g) is a pair of Taylor coefficients representing `Z_g`
  defined in [CD].
- cheng_tdT(g) is {\tilde T}_g as defined in [CD].

Basic data:

- _cheng_Phig_(...) is a helper function to compute
  `\mathcal{E}(\Phi_g)` as defined in Section 2.2 of [Ra].
- cheng_Phig(g) is function computing `\mathcal{E}(\Phi_g)`.
- borcherds_lift(level, weight, n, order) computes a dictionary
  of maps representing Borcherds lifts.

Solving and final computations:

- cheng_solve_case(g) represent `\Phi_g` as a root of a product of
  rescaled Borcherds products.
- cheng_find_lift_by_linear_equation(g, conjectured_Ng) try to represent
  `\Phi_g` as a product of rescaled Borcherds products of
  level ``conjectured_Ng`` as described in Section 5.3 of [Ra].
- compute_borcherds_lift_data(sol) computes `k_g`, the exponents
  `e_{q_1}`, `e_\zeta`, and `e_{q_2}`, and the initial coefficients
  of all `\phi_{g, i}_\frakc`.
"""

from eta_products.eta_products import *

from modular_form_transformations import modular_forms_pivots
from modular_form_transformations__store import ( load_inclusion_matrix_level, 
                                                  load_inclusion_matrix_order,
                                                  load_order_isomorphism_matrix,
                                                  load_projected_cusp_expansion_matrices )

################################################################################
### Cheng's and Duncan's modular forms
################################################################################

cheng_Ng = { '1A'    : 1,
             '2A'    : 2,
             '2B'    : 4,
             '3A'    : 3,
             '3B'    : 9,
             '4A'    : 8,
             '4B'    : 4,
             '4C'    : 16,
             '5A'    : 5,
             '6A'    : 6,
             '6B'    : 36,
             '7AB'   : 7,
             '8A'    : 8,
             '10A'   : 20,
             '11A'   : 11,
             '12A'   : 24,
             '12B'   : 144,
             '14AB'  : 14,
             '15AB'  : 15,
             '21AB'  : 63,
             '23AB'  : 23 }

cheng_tdT_level =  { '1A'    : 1,
                     '2A'    : 2,
                     '2B'    : 4,
                     '3A'    : 3,
                     '3B'    : 3,
                     '4A'    : 8,
                     '4B'    : 4,
                     '4C'    : 4,
                     '5A'    : 5,
                     '6A'    : 6,
                     '6B'    : 6,
                     '7AB'   : 7,
                     '8A'    : 8,
                     '10A'   : 10,
                     '11A'   : 11,
                     '12A'   : 12,
                     '12B'   : 12,
                     '14AB'  : 14,
                     '15AB'  : 15,
                     '21AB'  : 21,
                     '23AB'  : 23 }

cheng_chi = { '1A'    : 24,
              '2A'    : 8,
              '2B'    : 0,
              '3A'    : 6,
              '3B'    : 0,
              '4A'    : 0,
              '4B'    : 4,
              '4C'    : 0,
              '5A'    : 4,
              '6A'    : 2,
              '6B'    : 0,
              '7AB'   : 3,
              '8A'    : 2,
              '10A'   : 0,
              '11A'   : 2,
              '12A'   : 0,
              '12B'   : 0,
              '14AB'  : 1,
              '15AB'  : 1,
              '21AB'  : 0,
              '23AB'  : 1 }

### The cycle structure of powers of conjugacy classes `g` in `M_24`.
### The values of this dictionaries are pairs `(n_g, C_g)`, where `n_g`
### is the exponent of `g` as in [CD], and `C_g` is a dictionary, whose
### keys are the divisors `n` of `n_g`, and the values are the label
### of the conjugacy class `g^n`.

cheng_cycle_powers = \
    { '1A'   : (1,  { 1 : '1A' }),
      '2A'   : (2,  { 1 : '2A',    2 : '1A' }),
      '2B'   : (2,  { 1 : '2B',    2 : '1A' }),
      '3A'   : (3,  { 1 : '3A',    3 : '1A' }),
      '3B'   : (3,  { 1 : '3B',    3 : '1A' }),
      '4A'   : (4,  { 1 : '4A',    2 : '2A',  4 : '1A' }),
      '4B'   : (4,  { 1 : '4B',    2 : '2A',  4 : '1A' }),
      '4C'   : (4,  { 1 : '4C',    2 : '2B',  4 : '1A' }),
      '5A'   : (5,  { 1 : '5A',    5 : '1A' }),
      '6A'   : (6,  { 1 : '6A',    2 : '3A',  3 : '2A',  6 : '1A' }),
      '6B'   : (6,  { 1 : '6B',    2 : '3B',  3 : '2B',  6 : '1A' }),
      '7AB'  : (7,  { 1 : '7AB',   7 : '1A' }),
      '8A'   : (8,  { 1 : '8A',    2 : '4A',  4 : '2A',  8 : '1A' }),
      '10A'  : (10, { 1 : '10A',   2 : '5A',  5 : '2B', 10 : '1A' }),
      '11A'  : (11, { 1 : '11A',  11 : '1A' }),
      '12A'  : (12, { 1 : '12A',   2 : '6A',  3 : '4A',  4 : '3A',
                      6 : '2A',   12 : '1A' }),
      '12B'  : (12, { 1 : '12B',   2 : '6B',  3 : '4C',  4 : '3B',
                      6 : '2B',   12 : '1A' }),
      '14AB' : (14, { 1 : '14AB',  2 : '7AB', 7 : '2A', 14 : '1A' }),
      '15AB' : (15, { 1 : '15AB',  3 : '5A',  5 : '3A', 15 : '1A' }),
      '21AB' : (21, { 1 : '21AB',  3 : '7AB', 7 : '3B', 21 : '1A' }),
      '23AB' : (23, { 1 : '23AB', 23 : '1A' }) }

### Basic functions defined in [CD]

### cheng_Zg is the `M_24` twisted elliptic genus, represented
### by a pair of Taylor coefficients `tc_1, tc_2`.

def cheng_Zg(g) :
    r"""
    A pair of sums of extended eta products. The first is `Z_g(\tau, 0)`,
    the second is the corrected second derivative of `Z_g(\tau, z)` at `z = 0`.
    """
    # the first corrected Taylor coefficient of `\phi_{0,1}` is `12`,
    # the second is zero.
    tc1 = XEtaProductAlgebra(1)(12 * cheng_chi[g] / Integer(12))

    # the first corrected Taylor coefficient of `\phi_{-2, 1}` is zero,
    # the second is `1`.
    tc2 = cheng_tdT(g)

    return (tc1, tc2)

def cheng_tdT(g) :
    if g == '1A' :
        return XEtaProductAlgebra(1).zero()
    if g == '2A' :
        return 16 * eta_lambda(2)
    if g == '2B' :
        return -24 * eta_lambda(2) + 8 * eta_lambda(4)
    if g == '3A' :
        return 6 * eta_lambda(3)
    if g == '3B' :
        return 2 * eta_shift(1)**6 * etainv_shift(3)**2
    if g == '4A' :
        return 4 * eta_lambda(2) - 6 * eta_lambda(4) + 2 * eta_lambda(8)
    if g == '4B' :
        return 4 * (- eta_lambda(2) + eta_lambda(4))
    if g == '4C' :
        return 2 * eta_shift(1)**4 * eta_shift(2)**2 * etainv_shift(4)**2
    if g == '5A' :
        return 2 * eta_lambda(5)
    if g == '6A' :
        return 2 * (- eta_lambda(2) - eta_lambda(3) + eta_lambda(6))
    if g == '6B' :
        return 2 * eta_shift(1)**2 * eta_shift(2)**2 * eta_shift(3)**2 \
                 * etainv_shift(6)**2
    if g == '7AB' :
        return eta_lambda(7)
    if g == '8A' :
        return - eta_lambda(4) + eta_lambda(8)
    if g == '10A' :
        return 2 * eta_shift(1)**3 * eta_shift(2) * eta_shift(5) \
                 * etainv_shift(10)
    if g == '11A' :
        return Integer(2)/5 * (  eta_lambda(11)
                               - 11 * eta_shift(1)**2 * eta_shift(11)**2 )
    if g == '12A' :
        return 2 * eta_shift(3)**3 * eta_shift(4)**2 * eta_shift(6)**3 \
                 * etainv_shift(2) * etainv_shift(3) * etainv_shift(12)**2
    if g == '12B' :
        return 2 * eta_shift(1)**4 * eta_shift(4) * eta_shift(6) \
                 * etainv_shift(2) * etainv_shift(12)
    if g == '14AB' :
        return Integer(1)/3 * \
            ( - eta_lambda(2) - eta_lambda(7) + eta_lambda(14)
              - 14 * eta_shift(1) * eta_shift(2) * eta_shift(7) * eta_shift(14) )
    if g == '15AB' :
        return Integer(1)/4 * \
            ( - eta_lambda(3) - eta_lambda(5) + eta_lambda(15)
              - 15 * eta_shift(1) * eta_shift(3) * eta_shift(5) * eta_shift(15) )
    if g == '21AB' :
        return Integer(1)/3 * \
            ( 7 * eta_shift(1)**3 * eta_shift(7)**3
                * etainv_shift(3) * etainv_shift(21) 
              - eta_shift(1)**6 * etainv_shift(3)**2)
    if g == '23AB' :
        return Integer(1)/11 * (   eta_lambda(23)
                                 - 23 * eta_shift(1)**2 * eta_shift(23)**2
                                 + 23 * XEtaProductAlgebra(23)( 
                                               eta_product((3,-1,3,-1), 46)
                                         + 4 * eta_product((1,1,1,1), 46)
                                         + 4 * eta_product((0,2,0,2), 46) ) )

    raise ValueError( "Unknown value of g(={0})".format(g) )

### Jacobi forms (of weight `0` and index `1` given by the first two
###               corrected Taylor coefficients)

def _cheng_Phig_(M, cycle_powers, cheng_Zg) :
    r"""
    The Jacobi forms that give the product expansion of `\Phi_g`.
    See [BR] for a detailed explanation.
    
    INPUT:
    
    - `M`              -- An integer; The exponent of `g`.
    
    - ``cycle_powers`` -- A dictionary that encodes the cycle shape of
                          `g^d` for `d \isdiv M`. See ``cheng_cycle_powers``.
    
    - ``cheng_Zg``     -- The Jacobi forms `Z_g` defined in [CD].

    OUTPUT:
    
    - A dictionary mapping divisors of `M` to to pairs of extended eta products.
      The latter represent the first two corrected Taylor coefficients of a
      Jacobi form of index `1`.
    """
    res = dict()
    for d in divisors(M) :
        Zgd1 = XEtaProductAlgebra().zero()
        Zgd2 = XEtaProductAlgebra().zero()
        
        for (sd, gsd) in cycle_powers.iteritems() :
            if not Integer(sd).divides(d) or not Integer(d/sd).is_squarefree() :
                continue

            (Zgsd1, Zgsd2) = cheng_Zg(gsd)

            Zgsd1 *= (-(-1)**len(Integer(d/sd).divisors()))
            Zgsd2 *= (-(-1)**len(Integer(d/sd).divisors()))

            (Zgd1, Zgd2) = (Zgd1 + Zgsd1, Zgd2 + Zgsd2)

        res[d] = (Zgd1 * (1 / d), Zgd2 * (1 / d))

    return res

try :
    _tmp = _cheng_Phig
except :
    _cheng_Phig = dict()
def cheng_Phig(g) :
    r"""
    The Jacobi forms that give the product expansion of `\Phi_g`.

    OUTPUT:

    - See ``_cheng_Phig_``.
    """
    global _cheng_Phig

    try :
        return copy(_cheng_Phig[g])
    except KeyError :
        pass

    (M, cycle_powers) = cheng_cycle_powers[g]

    ## Representations of cheng_Phig in terms of eta products
    cheng_Phig_eta = _cheng_Phig_(M, cycle_powers, cheng_Zg)

    cheng_Phig_dict = dict()

    for (d, (tc0, tc2)) in cheng_Phig_eta.iteritems() :
        tc0 = vector(QQ, [tc0.eval_projection(1)[0]])
        
        pivots = modular_forms_pivots( cheng_tdT_level[g], 2 )
        prec = max(pivots) + 1
        tc2 = tc2.eval_projection(prec)
        tc2 = vector(QQ, [tc2[i] for i in pivots])
        
        cheng_Phig_dict[d] = (tc0, tc2)


    _cheng_Phig[g] = cheng_Phig_dict

    return copy(cheng_Phig_dict)

################################################################################
## Borcherds lift
################################################################################

def borcherds_lift( level, weight, n, order = 0 ) :
    r"""
    Given a level and a scaling factor `n` compute the dictionary of maps p_d of modular forms
    with the property:

    - Fix a `f` is a weakly holomorphic modular form of given pole
      order and level.  Let `v` be the coordinate vector associated to
      `f` with respect to the echelon basis of the ambient space. Then
      `\cE(B_N[f, n])(d) = p_d v`, where `N` is the given level.

    """
    proj_matrices = load_projected_cusp_expansion_matrices( level, weight, order )

    G = Gamma0(level)
    G_cusps = G.cusps()
    G_width = [Integer(G.cusp_data(c)[1]) for c in G_cusps]
    G_scalings = [level // gcd(level, c.denominator()) for c in G_cusps]

    lift = dict()
    for (c,w,s) in zip(G_cusps, G_width, G_scalings) :
        proj_matrix = proj_matrices(c)
        try :
            lift[n * s] += (w / s) * proj_matrix
        except KeyError :
            lift[n * s] = (w / s) * proj_matrix

    return lift


### The search for a rescaled Borcherds products.

def cheng_solve_case(g, conjectured_Ng = None, max_phi_level = None, order = 0, minimize_scaling = False) :
    r"""
    Try to solve the case `g`, which is the label of a conjugacy class of `M_{24}` by successifly
    increasing the level `N'_g`.  The quotient of `N'_g` and the level of lifted
    forms is at most ``additional_Ng``.

    INPUT:
    
    - `g`               -- A string.
    
    - ``conjectured_Ng`` -- A positive integer (default: `N_g`).  Try levels
                          `N'_g` that are greater than or equal to this.

    - ``max_phi_level`` -- A positive integer (default: level of `\tilde{T}_g`).
                           Lift weak Jacobi forms of at most this level.

    - ``order``     -- A non-negative integer (default: `1`). Maximal
                       pole order at infinity of lifted Jacobi forms lifted.
                           
    - ``minimize_scaling`` -- A boolean (default: ``True``).  If ``True`` the scaling factors
                              that occur in the resulting solution will be minimal.
    """
    if order < 0 :
        raise ValueError( "pole order must be non-negative." )
    if conjectured_Ng is None :
        conjectured_Ng = cheng_Ng[g]
    conjectured_Ng = Integer(conjectured_Ng)

    if max_phi_level is None :
        max_phi_level = cheng_tdT_level[g]
    max_phi_level = Integer(max_phi_level)

    if max_phi_level % cheng_tdT(g).parent().level() != 0 :
        raise ValueError( "tdT must be contained in considered space of weak Jacobi forms" )
    if conjectured_Ng % max_phi_level != 0 :
        raise ValueError( "Corrected value of Ng must be divisible by level of lifted weak Jacobi forms" )

    # Each time we are not able to prove that \Phi_g is modular of level
    # conjectured_Ng, we increase it by the factor Ng_base.
    Ng_base = prod( Integer(conjectured_Ng).prime_divisors() )

    while True :
        # Try to represent \Phi_g as a product of rescaled Borcherds
        # products.  If this failes, increase conjectured_Ng.
        try :
            (sol_vec, column_labels) = cheng_find_solution_by_linear_equation(g, conjectured_Ng, max_phi_level, order, minimize_scaling)
        except ValueError, e :
            if e.message != "matrix equation has no solutions" :
                raise
            
            conjectured_Ng *= Ng_base
            continue
        
        # save the result for later analysis.
        sol = (g, conjectured_Ng, max_phi_level, sol_vec, column_labels)
        save( sol, "cheng_results/sol_{0}.sobj".format(g) )
        return sol

def cheng_find_solution_by_linear_equation(g, conjectured_Ng, max_phi_level, order, minimize_scaling) :
    r"""
    Try to represent `\Phi_g` as a product of rescaled Borcherds products of
    level ``conjectured_Ng`` as described in Section 5 of [Ra].  Only
    rescaled Borcherds products coming from weak Jacobi forms of level
    ``max_phi_level`` are taken into consideration.

    We assume that ``max_phi_level`` divides ``conjectured_Ng``.

    INPUT:

    - `g`               -- A string.
    
    - ``conjectured_Ng`` -- A positive integer, instance of ``Integer``.

    - ``max_phi_level`` -- A positive integer, instance of ``Integer``.

    - ``order``     -- A non-negative integer (default: `1`). Maximal
                       pole order at infinity of lifted Jacobi forms lifted.

    - ``minimize_scaling`` -- A boolean. If ``True`` the scaling factors
                              that occur in the resulting solution will be minimal.

    OUTPUT:

    - A pair ``(sol_vec, coordinate_labels)`` of a solution vector whose
      coordinates are specified by ``coordinate_labels``.
      ``coordinate_labels`` is a list that associates to each coordinate
      a triple ``(level, scaling, i)``.  This means that the the corresponding
      components of ``sol_vec`` corresponds to a rescaled Borcherds product
      with given scaling  of a weakly holomorphic modular form of given level.
      The last entry, `i`, refers to the `i`th vector in the concatenation of
      the echelon bases for weakly holomorphic modular forms of weight
      `0` and `2` whose pole order is bounded by ``order``.

    """
    square_free_level = max_phi_level.is_squarefree()

    nmb_components = len(conjectured_Ng.divisors())
    mf_wt0_dim = dict( ( level, 
                         ModularForms(level, 12 * order).dimension() \
                             if order != 0 else 1 )
                       for level in max_phi_level.divisors() )
    mf_wt2_dim = dict( ( level,
                         ModularForms(level, 2 + 12 * order).dimension() )
                        for level in max_phi_level.divisors() )


    ## We start by computing the vector that corresponds to Cheng's
    ## and Duncan's forms.

    ## This is `\mathcal{E}( \Phi_g )`.
    cheng_Phig_dict = cheng_Phig(g)
    if max_phi_level != cheng_tdT_level[g] :
        for d in cheng_Phig_dict.keys() :
            (tc0, tc2) = cheng_Phig_dict[d]
            cheng_Phig_dict[d] = \
                ( tc0,
                  load_inclusion_matrix_level(cheng_tdT_level[g], max_phi_level, 2)
                  * tc2 )
    if order != 0 :
        for d in cheng_Phig_dict.keys() :
            (tc0, tc2) = cheng_Phig_dict[d]
            cheng_Phig_dict[d] = \
                ( load_inclusion_matrix_order( max_phi_level, 0, 0, order )
                    * tc0,
                  load_inclusion_matrix_order( max_phi_level, 2, 0, order )
                    * tc2 )

    tc0_list = list()
    tc2_list = list()
    for d in conjectured_Ng.divisors() :
        try :
            tc0_list += list(cheng_Phig_dict[d][0])
            tc2_list += list(cheng_Phig_dict[d][1])
        except KeyError :
            tc0_list += mf_wt0_dim[max_phi_level] * [0]
            tc2_list += mf_wt2_dim[max_phi_level] * [0]
    cheng_Phig_vec = vector(QQ, tc0_list + tc2_list)


    ## We compute the span of Borcherds products as the column space
    ## of the matrix cE_bor_mat
    cE_bor_mat = zero_matrix( QQ if square_free_level
                                 else UniversalCyclotomicField(),
                              nmb_components * (mf_wt0_dim[max_phi_level] + mf_wt2_dim[max_phi_level]), 0)
    ## column_labels associate to each column of cE_bor_mat tuple
    ## ``(level, scaling, i)``, where `i` refers to the `i`th vector
    ## in the concatenation of the echelon bases for weakly holomorphic
    ## modular forms of weight `0` and `2`.
    column_labels = list()
    if minimize_scaling :
        if square_free_level :
            cE_bor_mat_span = cE_bor_mat.column_module()
    for scaling in conjectured_Ng.divisors() :
        for cur_phi_level in gcd(max_phi_level, conjectured_Ng // scaling).divisors() :
            cE_bor_tc0 = borcherds_lift( cur_phi_level, 0, scaling, order )
            cE_bor_tc2 = borcherds_lift( cur_phi_level, 2, scaling, order )

            if cur_phi_level != max_phi_level :
                for d in cE_bor_tc0.keys() :
                    cE_bor_tc0[d] = \
                        load_inclusion_matrix_level(cur_phi_level, max_phi_level, 0, order) \
                        * cE_bor_tc0[d]
                    cE_bor_tc2[d] = \
                        load_inclusion_matrix_level(cur_phi_level, max_phi_level, 2, order) \
                        * cE_bor_tc2[d]

            cE_bor_this = zero_matrix( QQ if square_free_level
                                          else UniversalCyclotomicField(),
                                       nmb_components * (mf_wt0_dim[max_phi_level] + mf_wt2_dim[max_phi_level]),
                                       mf_wt0_dim[cur_phi_level] + mf_wt2_dim[cur_phi_level] )

            for (i,d) in enumerate(conjectured_Ng.divisors()) :
                try :
                    cE_bor_this.set_block( i * mf_wt0_dim[max_phi_level],
                                           0,
                                           cE_bor_tc0[d] )
                    cE_bor_this.set_block( nmb_components * mf_wt0_dim[max_phi_level] + i * mf_wt2_dim[max_phi_level],
                                           mf_wt0_dim[cur_phi_level],
                                           cE_bor_tc2[d] )
                except KeyError :
                    pass

            if not minimize_scaling :
                cE_bor_mat = cE_bor_mat.augment( cE_bor_this )
                
                for i in range(mf_wt0_dim[cur_phi_level] + mf_wt2_dim[cur_phi_level]) :
                    column_labels.append( (cur_phi_level, scaling, i) )
            else :
                if square_free_level :
                    for (i, col) in enumerate(cE_bor_this.columns()) :
                        if col not in cE_bor_mat_span :
                            column_labels.append( (cur_phi_level, scaling, i) )
                            cE_bor_mat = cE_bor_mat.augment(col)
                            cE_bor_mat_span += span([col])
                else :
                    cyclotomic_order = lcm([ e.field_order() for e in cE_bor_mat.list() + cE_bor_this.list() ])
                    K = CyclotomicField( cyclotomic_order )
                    coord_func = K.gen().coordinates_in_terms_of_powers()
                    cE_bor_mat_coords = flatten(zip(*[ coord_func(K(e.to_cyclotomic_field())) for e in cE_bor_mat.list() ]))

                    cE_bor_mat_span = matrix( QQ, K.degree() * cE_bor_mat.nrows(),
                                              cE_bor_mat_coords ).column_space()

                    for (i, col) in enumerate(cE_bor_this.columns()) :
                        col_QQ = vector( QQ, flatten(zip(*[ coord_func(e.to_cyclotomic_field()) for e in col ])) )

                        if col_QQ not in cE_bor_mat_span :
                            column_labels.append( (cur_phi_level, scaling, i) )
                            cE_bor_mat = cE_bor_mat.augment(col)
                            cE_bor_mat_span += span([col_QQ])


    if square_free_level :
        return (cE_bor_mat.solve_right(cheng_Phig_vec), column_labels)
    else :
        cyclotomic_order = lcm([ e.field_order() for e in cE_bor_mat.list() ])
        K = CyclotomicField( cyclotomic_order )
        coord_func = K.gen().coordinates_in_terms_of_powers()
        cE_bor_mat_coords = zip(*[ coord_func(K(e.to_cyclotomic_field())) for e in cE_bor_mat.list() ])

        cyclotomic_vanishing_space = FreeModule(QQ, cE_bor_mat.ncols())
        for cE_bor_mat_coords__non_QQ in cE_bor_mat_coords[1:] :
            cyclotomic_vanishing_space = \
                cyclotomic_vanishing_space.intersection(
                matrix( QQ, cE_bor_mat.nrows(), cE_bor_mat_coords__non_QQ).right_kernel() )

        cyclotomic_vanishing_basis = cyclotomic_vanishing_space.basis_matrix().transpose()
        cE_bor_mat_QQ = matrix( QQ, cE_bor_mat.nrows(), cE_bor_mat_coords[0] )
        sol = (cE_bor_mat_QQ * cyclotomic_vanishing_basis).solve_right(cheng_Phig_vec)

        return (cyclotomic_vanishing_basis * sol, column_labels)

def compute_borcherds_lift_data(sol) :
    r"""
    Given a representation of `\Phi_g` as a rescaled Borcherds products
    `\prod_i B[\phi_i, n_i]` as stored by ``cheng_solve_case``,
    compute:

    -  The weight `k_g` of `\Phi_g`.

    -  The exponents `e_{q_1}` etc. of `q_1`, `\zeta`, and `q_2`.

    -  The coefficients of q^0 \zeta^0 and q^0 \zeta^1 of all `\phi_i`.
       From this, one can infer `p_g` as described in Section 5.1 of [Ra].

    See Section 5.3 of [Ra].
    
    INPUT:
    
    - ``sol`` -- Content of a solution file.
                 A tuple of ``(string, integer, integer, vector of rationals, column_labels)``. 

    EXAMPLES:
    
        sage: compute_borcherds_lift_data(load('cheng_results/sol_2A.sobj'))
        (6, (1, 1, 1))
    """
    max_phi_level = sol[2]
    sol_vec = sol[3]
    column_labels = sol[4]

    proj_matrices = dict( (level, load_projected_cusp_expansion_matrices(level, 2))
                           for level in max_phi_level.divisors() )

    ## We are reduced to applying the formula for the weight
    ## and for `e_{q_1}`, etc. in a straight forward way.
    k = 0
    (n, r, m) = (0, 0, 0)
    initial_coefficients = dict()
    for (e, (level, scaling, coord)) in zip(sol_vec, column_labels) :
        ## if e == 0, then the corresponding lift does not contribute anything.
        if e == 0 :
            continue

        ## Assert that the coefficient `e` an element of `\QQ`
        ## (and not an instance of ``int``).
        e = QQ(e)

        ## Here we pick pre-factors for `k`, `e_{q_1}`, etc.  These
        ## factors correspond to the Fourier expansion of `\phi_{0, 1} / 12`
        ## and `\phi_{-2, 1}`.
        if coord == 0 :
            k_factor = QQ((10, 12))
            n_factor = 1
            r_factor = QQ((1,12))
            m_factor = QQ((1,6))
        else :
            k_factor = -2
            n_factor = 0
            r_factor = QQ(1)
            m_factor = 2

        ## Here we pick factors for the Fourier coefficient of `q^0 \zeta^0`
        ## and `q^0 \zeta^1`. These factors correspond to the Fourier
        ## expansion of `\phi_{0, 1} / 12` and `\phi_{-2, 1}`.
        if coord == 0 :
            c0_factor = QQ((10,12)) #-QQ((1,6))
            c1_factor = QQ((1,12))
        else :
            c0_factor = -2
            c1_factor = 1

        ## The next statements all come from the formulas in Section 3.
        G = Gamma0(level)
        for cusp in G.cusps() :
            width = G.cusp_data(cusp)[1]
            denominator = gcd(level, cusp.denominator())
            width_quotient = width / (level / denominator)

            if coord != 0 :
                cusp_factor = proj_matrices[level](cusp)[0, coord - 1]
            else :
                cusp_factor = 1

            k += e * (Integer(1)/2) * cusp_factor * k_factor * width_quotient

            n += e * (Integer(1)/24) * cusp_factor * scaling * width * n_factor
            r += e * (Integer(1)/2) * cusp_factor * scaling * width * r_factor
            m += e * (Integer(1)/4) * cusp_factor * scaling * width * m_factor

            try :
                (c0, c1) = initial_coefficients[(level, scaling, cusp)]
            except KeyError :
                (c0, c1) = (0, 0)

            c0 += e * cusp_factor * c0_factor * width_quotient
            c1 += e * cusp_factor * c1_factor * width_quotient

            initial_coefficients[(level, scaling, cusp)] = (c0, c1) 

    
    return (QQ(k), (QQ(n), QQ(r), QQ(m)), initial_coefficients)
