import unittest


from sage.all import *

from eta_products import *
from eta_products import (_divide_q_power, _eta, _eta_d,
                          _zeta_denominator_xep, _eval_xep,
                          _eval_xep__complex,
                          _eval_projection_xep, _cusp_matrix,
                          _cusp_expansion, _eta_multiplier,
                          _square_normalization_xep)


class TestXEtaProduct_class(unittest.TestCase) :

    def setUp(self) :
        pass

    def test___init__(self) :
        self.assertIs( XEtaProduct_class( ((0,2,1,0),), None, False ).parent(),
                       XEtaProducts() )

        class dXEtaProducts( XEtaProducts ) :
            pass
        self.assertIs( XEtaProduct_class( ((0,2,1,0),), dXEtaProducts(), False ).parent(),
                       dXEtaProducts() )

        # do not sort
        self.assertEqual( XEtaProduct_class( ((1,1,1,0),(-1,1,1,0)), clean = False )._e,
                          ((1,1,1,0), (-1,1,1,0)) )

        # do not oversimplify
        self.assertEqual( XEtaProduct_class( ((1,1,1,0),(-1,3,1,0)) )._e,
                          ((-1,3,1,0), (1,1,1,0)) )
        self.assertEqual( XEtaProduct_class( ((1,1,1,0),(-1,1,3,0)) )._e,
                          ((-1,1,3,0), (1,1,1,0)) )
        self.assertEqual( XEtaProduct_class( ((1,1,1,0),(-1,1,1,2)) )._e,
                          ((-1,1,1,2), (1,1,1,0)) )

        # sort entries
        self.assertEqual( XEtaProduct_class( ((1,1,1,0),(-1,1,3,0)) )._e,
                          ((-1,1,3,0), (1,1,1,0)) )
        self.assertEqual( XEtaProduct_class( ((1,1,3,0),(1,1,1,0)) )._e,
                          ((1,1,1,0), (1,1,3,0)) )

        # cancel eta functions
        self.assertEqual( XEtaProduct_class( ((1,1,1,0),(-1,1,1,0)) )._e,
                          tuple() )
        self.assertEqual( XEtaProduct_class( ((-1,1,1,0),(1,1,1,0)) )._e,
                          tuple() )
        self.assertEqual( XEtaProduct_class( ((1,1,1,0),(0,3,1,0),(-1,1,1,0)) )._e,
                          ((0,3,1,0),) )


    def test__maximal_q_denominator(self) :
        eps = XEtaProducts()

        self.assertEqual( eps((0,3,1,0))._maximal_q_denominator(),
                          1 )
        self.assertEqual( eps((0,1,3,0))._maximal_q_denominator(),
                          3 )
        self.assertEqual( eps((0,7,5,2))._maximal_q_denominator(),
                          5 )

        self.assertEqual( eps((1,1,1,0))._maximal_q_denominator(),
                          24 )
        self.assertEqual( eps((1,1,1,3))._maximal_q_denominator(),
                          24 )
        self.assertEqual( eps((1,3,1,0))._maximal_q_denominator(),
                          8 )
        self.assertEqual( eps((1,1,2,0))._maximal_q_denominator(),
                          48 )

    def test__zeta_denominator(self) :
        eps = XEtaProducts()

        self.assertEqual( eps((0,3,1,0))._zeta_denominator(),
                          1 )
        self.assertEqual( eps((1,3,1,0))._zeta_denominator(),
                          1 )
        self.assertEqual( eps((1,3,1,1))._zeta_denominator(),
                          24 )
        self.assertEqual( eps((0,1,2,1))._zeta_denominator(),
                          2 )
        self.assertEqual( eps((-1,1,3,2))._zeta_denominator(),
                          36 )

    def test__square_normalization(self) :
        eps = XEtaProducts()
        M = Integer(randint(1, 200))
        N = Integer(randint(1, 200))
        h = Integer(randint(-200, 200))

        self.assertEqual( eps((1,M,N,h))._square_normalization(), 
                          1/N )
        self.assertEqual( (eps((1,M,N,h))**5)._square_normalization(),
                          1/N**5 )
        self.assertEqual( (eps((1,M,N,h))**5 * eps((-1, 1, N, 1)))._square_normalization(),
                          1/N**4 )

    def test_weight(self) :
        eps = XEtaProducts()

        self.assertEqual( eps((1,1,1,0)).weight(),
                          Integer(1)/2 )
        self.assertEqual( eps((-1,1,1,0)).weight(),
                          -Integer(1)/2 )
        self.assertEqual( eps((1,1,1,3)).weight(),
                          Integer(1)/2 )
        self.assertEqual( (eps((1,1,1,0))**3).weight(),
                          Integer(3)/2 )
        self.assertEqual( (eps((1,1,1,0)) * eps((-1,3,1,0))).weight(),
                          0 )

        self.assertEqual( eps((0,2,1,0)).weight(),
                          2 )
        self.assertEqual( (eps((0,2,1,0)) * eps((1,1,1,0))).weight(),
                          Integer(5)/2 )

    def test_rescale(self) :
        eps = XEtaProducts()

        self.assertEqual( eps((1,1,1,3)).rescale(2), 
                          eps((1,2,1,3)) )
        self.assertEqual( eps((0,7,4,4)).rescale(7), 
                          eps((0,49,4,4)) )

    def test_eval_projection(self) :
        eps = XEtaProducts()
        
        self.assertRaises( ValueError,
                           lambda e: e.eval_projection(-1, QQ),
                           eps((0,2,1,0)) )

        self.assertEqual( eps((0,2,1,0)).eval_projection(20, QQ),
                          _eval_projection_xep(((0,2,1,0),), 20) )
        self.assertEqual( eps((0,1,2,0)).eval_projection(20, QQ),
                          _eval_projection_xep(((0,1,2,0),), 20) )
        self.assertEqual( eps((1,1,3,0)).eval_projection(20, CyclotomicField(3)),
                          0 )

    def test__mul_(self) :
        eps = XEtaProducts()

        self.assertEqual( eps((0,2,1,0)) * eps((0,3,1,0)),
                          eps([(0,2,1,0), (0,3,1,0)]) )
        self.assertEqual( (eps((1,3,1,0)) * eps((-1,2,1,0)))._e,
                          ((-1,2,1,0), (1,3,1,0)) )
        self.assertEqual( eps((1,2,3,7))**3,
                          eps(3*[(1,2,3,7)]) )

    def test___invert__(self) :
        eps = XEtaProducts()
        
        self.assertEqual( eps((1,2,3,7))**-1,
                          eps((-1,2,3,7)) )
        self.assertEqual( (eps((1,2,3,7)) * eps((-1,3,1,0)))**-1,
                          eps((-1,2,3,7)) * eps((1,3,1,0)) )

        self.assertRaises( ValueError,
                           lambda e: e**-1,
                           eps((0,3,1,0)) )
        self.assertRaises( ValueError,
                           lambda e: e**-1,
                           eps((1,2,4,0))**4 * eps((0,3,1,0)) )
    
    def test___cmp__(self) :
        eps = XEtaProducts()

        self.assertEqual( eps(1),
                          eps(1) )
        self.assertEqual( eps((1,2,3,7)),
                          eps((1,2,3,7)) )

        self.assertNotEqual( eps((1,2,3,7)),
                             eps((1,2,3,0)) )
        self.assertNotEqual( eps((1,2,3,7)),
                             eps((1,2,2,7)) )
        self.assertNotEqual( eps((1,2,3,7)),
                             eps((1,5,3,7)) )
        self.assertNotEqual( eps((1,2,3,7)),
                             eps((1,2,3,7))**2 )
        self.assertNotEqual( eps((1,2,3,7)),
                             eps((1,2,3,7)) * eps((1,2,4,7)) )
        self.assertNotEqual( eps((1,2,3,7)),
                             eps(1) )

    def test__repr_(self) :
        eps = XEtaProducts()

        self.assertEqual( repr(eps((0,2,3,7))),
                          "Lambda_{2,3,7}(z)" )
        self.assertEqual( repr(eps((1,2,3,7))),
                          "eta((2z + 7)/3)" )
        self.assertEqual( repr(eps((1,2,3,0))),
                          "eta(2z/3)" )
        self.assertEqual( repr(eps((1,2,1,7))),
                          "eta(2z + 7)" )

    def test___hash__(self) :
        eps = XEtaProducts()

        self.assertEqual( hash(eps((0,2,3,7))),
                          -6478914342297160413 )
        self.assertEqual( hash(eps((1,2,3,7))**4),
                          7986182315527228540 )

class TestXEtaProducts(unittest.TestCase) :

    def setUp(self) :
        pass

    def test__class(self) :
        self.assertIsNone( TestSuite(XEtaProducts).run() )

        self.assertEqual( XEtaProducts().category(),
                          Monoids() )
        self.assertIs( XEtaProducts(),
                       XEtaProducts() )

    def test___init__(self) :
        self.assertIsInstance( XEtaProducts(),
                               XEtaProducts )

    def test_an_element(self) :
        self.assertIn( XEtaProducts().an_element(),
                       XEtaProducts() )

    def test__element_constructor_(self) :
        self.assertIn( XEtaProducts()(1),
                       XEtaProducts() )
        self.assertIn( XEtaProducts()((0,1,2,0)),
                       XEtaProducts() )
        self.assertEqual( XEtaProducts()(((0,1,2,0),)),
                          XEtaProducts()((0,1,2,0)) )
        self.assertEqual( XEtaProducts()([(0,1,2,0)]),
                          XEtaProducts()((0,1,2,0)) )
        self.assertIn( XEtaProducts()(((0,1,2,0), (1,4,7,2))),
                       XEtaProducts() )
        self.assertEqual( XEtaProducts()(((0,1,2,0), (1,4,7,2))),
                          XEtaProducts()([(0,1,2,0), (1,4,7,2)]) )

        self.assertRaises( TypeError,
                           XEtaProducts,
                           2 )
        self.assertRaises( TypeError,
                           XEtaProducts,
                           [(0,1,2,0), (1,4,7,2), (1,)] )
        self.assertRaises( TypeError,
                           XEtaProducts,
                           [(0,1,2,0), (2,4,7,2)] )
        self.assertRaises( TypeError,
                           XEtaProducts,
                           [(0,1,2,0), (1,Integer(1)/4,7,2)] )
        
    def test__repr_(self) :
        self.assertEqual( repr(XEtaProducts()),
                          "Set of all extended eta products" )

class TestXEtaProductSum(unittest.TestCase) :

    def setUp(self) :
        pass

    def test___init__(self) :
        eps = XEtaProducts()

        self.assertIs( XEtaProductSum( {eps(1): 1}, XEtaProductAlgebra(QQ) ).parent(),
                       XEtaProductAlgebra(QQ) )
        d = {eps(1): 1}
        self.assertIsNot( XEtaProductSum( {eps(1): 1}, XEtaProductAlgebra(QQ) )._dict,
                          d )
        
        self.assertEqual( XEtaProductSum( {eps((0,2,1,0)): Integer(1), eps((0,3,1,0)): 0},
                                          XEtaProductAlgebra(QQ) )._dict,
                          {eps((0,2,1,0)): 1} )
        self.assertEqual( XEtaProductSum( {eps((0,2,1,0)): QQ(1), eps((0,3,1,0)): 0},
                                          XEtaProductAlgebra(QQ), clean = False )._dict,
                          {eps((0,2,1,0)): 1, eps((0,3,1,0)): 0} )
        self.assertEqual( XEtaProductSum( {eps((0,2,1,0)): QQ(1), eps((0,3,1,0)): CyclotomicField(7)(0)},
                                          XEtaProductAlgebra(QQ), check = True )._dict,
                          {eps((0,2,1,0)): 1} )
        self.assertRaises( ValueError,
                           XEtaProductSum,
                           {eps((0,2,1,0)): QQ(1), eps((0,3,1,0)): CyclotomicField(7)(0)},
                           XEtaProductAlgebra(QQ),
                           clean = False, check = True )

    def test__maximal_q_denominator(self) :
        eps = lambda e: XEtaProductAlgebra()(XEtaProducts()(e))

        self.assertEqual( eps((0,2,1,0))._maximal_q_denominator(),
                          1 )
        self.assertEqual( (eps((1,3,1,0)) + eps((1,2,1,0)))._maximal_q_denominator(),
                          24 )
        self.assertEqual( (eps((0,1,3,2)) + eps((0,1,7,1)))._maximal_q_denominator(),
                          21 )

    def test_weight(self) :
        eps = lambda e: XEtaProductAlgebra()(XEtaProducts()(e))

        self.assertEqual( eps((0,2,1,0)).weight(),
                          2 )
        self.assertEqual( (eps((0,2,1,0)) + eps((0,3,1,0))).weight(),
                          2 )
        self.assertEqual( ((eps((0,2,1,0)) + eps((0,3,1,0)))**2).weight(),
                          4 )
        self.assertEqual( (eps((1,3,1,0))**2 * eps((-1,2,1,2))).weight(),
                          Integer(1)/2 )
        self.assertEqual( (eps((1,3,1,0))**2 * eps((-1,2,1,2)) + eps((1,1,3,0))).weight(),
                          Integer(1)/2 )
        self.assertIsNone( (eps((1,2,3,0)) + eps((0,3,1,0))).weight() )

    def test__eval_projection__check_cache(self) :
        eps = lambda e: XEtaProductAlgebra()(XEtaProducts()(e))

        e = eps((0,2,1,0))
        self.assertRaises( ValueError,
                           lambda e: e._eval_projection__check_cache(10),
                           e )
        ev = e._eval_projection__primitive(10, QQ)
        self.assertIsNot( e._eval_projection__check_cache(10),
                          ev )
        self.assertEqual( e._eval_projection__check_cache(10),
                          ev )
        self.assertEqual( e._eval_projection__check_cache(8),
                          ev.truncate_powerseries(8) )
        self.assertEqual( e._eval_projection__check_cache(0),
                          0 )
        

        self.assertEqual( XEtaProductAlgebra().zero()._eval_projection__check_cache(10),
                          0 )
        self.assertEqual( XEtaProductAlgebra().zero()._eval_projection__check_cache(10).prec(),
                          10 )

        self.assertRaises( ValueError,
                           eps(1)._eval_projection__check_cache,
                           -1 )
        
    def test__eval_projection__primitive(self) :
        eps = lambda e: XEtaProductAlgebra()(XEtaProducts()(e))

        self.assertEqual( (  3 * XEtaProductAlgebra(2)(XEtaProducts()((0,2,1,0)))
                           + 2 * XEtaProductAlgebra(3)(XEtaProducts()((0,3,1,0)))) \
                              ._eval_projection__primitive(10, QQ),
                            3 * _eval_projection_xep(((0,2,1,0),), 10)
                          + 2 * _eval_projection_xep(((0,3,1,0),), 10) )
        self.assertEqual( (  eps((1,1,4,1))**2 * eps((-1,1,2,1))
                           + 3 * eps((1,1,9,1))**3 * eps((-1,1,3,1)) )._eval_projection__primitive(3, None).parent().base_ring(),
                          CyclotomicField(36) )
        self.assertEqual( eps((0,2,2,1))._eval_projection__primitive(3, None).parent().base_ring(),
                          RationalField() )
        self.assertEqual( eps((0,2,1,0))._eval_projection__primitive(3, CyclotomicField(4)).parent().base_ring(),
                          CyclotomicField(4) )


        self.assertRaises( ValueError,
                           (  eps((1,1,4,1))**2 * eps((-1,1,2,1))
                              + 3 * eps((1,1,9,1))**3 * eps((-1,1,3,1)) )._eval_projection__primitive,
                           3, CyclotomicField(4) )
        
        e = eps((0,2,1,0))
        ev = e._eval_projection__primitive(10, None)
        ev += 1
        self.assertNotEqual( e._eval_projection__primitive(10, None),
                             ev )
        
    def test_eval_projection(self) :
        eps = lambda e: XEtaProductAlgebra()(XEtaProducts()(e))
        eps4 = lambda e: XEtaProductAlgebra(4)(XEtaProducts()(e))
        eps9 = lambda e: XEtaProductAlgebra(9)(XEtaProducts()(e))

        self.assertEqual( (CyclotomicField(7).gen(0) * eps(1)).eval_projection(10),
                          CyclotomicField(7).gen(0) )
        self.assertEqual( (3 * eps4((0,2,1,0)) + 2 * eps9((0,3,1,0)))\
                             .eval_projection(20, QQ),
                            3 * _eval_projection_xep(((0,2,1,0),), 20)
                          + 2 * _eval_projection_xep(((0,3,1,0),), 20) )
        self.assertEqual( (  eps4((0,1,4,1))
                           + 3 * eps9((1,1,9,1))**6 * eps9((-1,1,3,1))**2 ).eval_projection(3).parent().base_ring(),
                          CyclotomicField(36) )
        self.assertEqual( eps((0,2,2,1)).eval_projection(3).parent().base_ring(),
                          RationalField() )

        self.assertRaises( NotImplementedError,
                           lambda e: e.eval_projection(10),
                           eps((0,3,1,0)) + eps((0,2,1,0))**2 )

        e = eps((0,2,1,0))
        ev = e.eval_projection(10, None)
        ev += 1
        self.assertNotEqual( e.eval_projection(10, None),
                             ev )

        def test_numerical(es, tau) :
            assert all( log != 0 for e in es._dict.keys()
                                 for (log, _, _, _) in e._e )
            CC = ComplexField(200)
            
            e_qexp = es.eval_projection(5000)
            e_eval_std = e_qexp.polynomial().subs(q = CC(exp(2 * pi * I * tau)))

            eta_qexp = _eta(5000)

            e_eval_alt = CC(0)
            for _ in range(es._maximal_q_denominator()) :
                tau += 1

                for (e, v) in es._dict.iteritems() :
                    e_eval_alt_intermediate = CC(v) / es._maximal_q_denominator()

                    for (log, M, N, h) in e._e :
                        e_tau = (M * tau + h) / N

                        eta_eval =   CC(exp(2 * pi * I * e_tau / 24)) \
                                   * eta_qexp.polynomial().subs(q = CC(exp(2 * pi * I * e_tau)))
                        e_eval_alt_intermediate *= eta_eval**log

                    e_eval_alt += e_eval_alt_intermediate 

            self.assertAlmostEqual( e_eval_std,
                                    e_eval_alt )

        for tau in [I, 2 * I, I + Integer(1)/2, I/5] :
            test_numerical(eps9((1,9,1,0))**6 * eps9((-1,3,1,0))**2, tau)
            test_numerical(eps9((1,3,3,1))**6 * eps9((-1,3,1,0))**2, tau)
            f23_2 = XEtaProductAlgebra(23)(
                eta_shift(1)**3 * eta_shift(23)**3 \
                    * etainv_shift(2) * etainv_shift(46) \
                + 4 * ( eta_shift(1) * eta_shift(2) * eta_shift(23) \
                        * eta_shift(46) ) \
                + 4 * ( eta_shift(2)**2 * eta_shift(46)**2 ) )
            test_numerical(f23_2, tau)

    def test_cusp_expansion(self) :
        CC = ComplexField(200)

        # e : An extended eta product
        # c : A cusp
        def test_e_c(e, c, places = None) :
            (a_g, b_g, c_g, d_g) = _cusp_matrix(c)
            gamma = lambda tau: (a_g * tau + b_g) / (c_g * tau + d_g)
            j = lambda tau: CC(c_g * tau + d_g)**-1

            ce = e.cusp_expansion(c)

            e_qexp = e.eval_projection(5000)
            ce_qexp = ce.eval_projection(5000)

            ce_width = ce._maximal_q_denominator()

            tau = SR('I') + 1/3
            while(True) :
                e_eval = CC(Integer(1) / ce_width) * \
                    sum( j(tau + n)**e.weight() * e_qexp.polynomial().subs(q = CC(exp(2 * pi * I * gamma(tau + n))))
                         for n in range(ce_width) )

                if abs(e_eval) > 10**-4:
                    break
                tau = tau / 2

            ce_eval = CC(ce_qexp.polynomial().subs(q = CC(exp(2 * pi * I * tau))))
            if places is None :
                self.assertAlmostEqual( e_eval,
                                        ce_eval )
            else :
                self.assertAlmostEqual( e_eval,
                                        ce_eval,
                                        places )

        test_e_c( XEtaProductAlgebra(2, QQ)(XEtaProducts()((0,2,1,0))),
                  Cusp(0) )
        test_e_c( XEtaProductAlgebra(2, QQ)(XEtaProducts()((0,2,1,0))),
                  Cusp(Integer(1)/2) )
        test_e_c( XEtaProductAlgebra(4, QQ)(XEtaProducts()((0,4,1,0))),
                  Cusp(Integer(1)/2) )
        test_e_c( XEtaProductAlgebra(1, QQ)(XEtaProducts()((1,1,1,0))**24),
                  Cusp(0) )
        test_e_c( XEtaProductAlgebra(9, QQ)(XEtaProducts()((1,9,1,0))**6 * XEtaProducts()((-1,3,1,0))**2),
                  Cusp(Integer(1)/3) )
        test_e_c(   CyclotomicField(12).gen(0)
                  * XEtaProductAlgebra(9, CyclotomicField(12))(XEtaProducts()((1,9,1,0))**6 * XEtaProducts()((-1,3,1,0))**2),
                  Cusp(Integer(1)/3) )
        test_e_c(   CyclotomicField(7).gen(0)
                  * XEtaProductAlgebra(9, CyclotomicField(7))(XEtaProducts()((1,9,1,0))**6 * XEtaProducts()((-1,3,1,0))**2),
                  Cusp(Integer(1)/3) )

        f23_2 = XEtaProductAlgebra(23)(
            eta_shift(1)**3 * eta_shift(23)**3 \
                * etainv_shift(2) * etainv_shift(46) \
            + 4 * ( eta_shift(1) * eta_shift(2) * eta_shift(23) \
                    * eta_shift(46) ) \
            + 4 * ( eta_shift(2)**2 * eta_shift(46)**2 ) )
        test_e_c(f23_2, Cusp(0), 5)

        test_e_c( eta_lambda(2), Cusp(0) )
        test_e_c( eta_lambda(4), Cusp(0) )
        test_e_c( eta_lambda(5), Cusp(0) )
        test_e_c( eta_lambda(10), Cusp(0) )
        test_e_c( eta_lambda(20), Cusp(0) )
        test_e_c( eta_product((2, 1, 0, 0, 0, 1), 20), Cusp(0) )
        

        self.assertRaises( NotImplementedError,
                           lambda e: e.cusp_expansion(Cusp(Integer(1)/3)),
                             QuadraticField(-7).gen() \
                           * XEtaProductAlgebra(9, QuadraticField(-7))(XEtaProducts()((1,3,1,0))) )
        self.assertRaises( NotImplementedError,
                           lambda e: e.cusp_expansion(Cusp(Integer(1)/3)),
                             CyclotomicField(7, embedding = CLF(exp(3 * 2 * pi * I / 7))).gen() \
                           * XEtaProductAlgebra(9, CyclotomicField(7, embedding = CLF(exp(3 * 2 * pi * I / 7))))(XEtaProducts()((1,3,1,0))) )

    def test__add_(self) :
        eps = lambda e: XEtaProductAlgebra(None, QQ)(XEtaProducts()(e))

        self.assertEqual( (eps((0,2,1,0)) + eps((0,2,1,0)))._dict,
                          { XEtaProducts()((0,2,1,0)) : 2 } )
        self.assertEqual( (eps((0,2,1,0)) + 2 * eps((0,2,4,0)))._dict,
                          { XEtaProducts()((0,2,1,0)) : 1,
                            XEtaProducts()((0,2,4,0)) : 2 } )

    def test__mul_(self) :
        eps = lambda e: XEtaProductAlgebra(None, QQ)(XEtaProducts()(e))

        self.assertEqual( (eps((0,2,1,0)) * eps((0,2,1,0)))._dict,
                          { XEtaProducts()((0,2,1,0))**2 : 1 } )
        self.assertEqual( (eps((0,2,1,0)) + 2 * eps((0,2,4,0)))._dict,
                          { XEtaProducts()((0,2,1,0)) : 1,
                            XEtaProducts()((0,2,4,0)) : 2 } )

    def test___cmp__(self) :
        eps = lambda e: XEtaProductAlgebra(None, QQ)(XEtaProducts()(e))
        
        self.assertEqual( eps((0,2,1,0)) + eps((1,3,1,4)),
                          eps((0,2,1,0)) + eps((1,3,1,4)) )
        self.assertNotEqual( eps((0,2,1,0)) + eps((1,3,1,4)),
                             eps((0,2,1,0)) + eps((1,3,1,3)) )

    def test__repr_(self) :
        eps = lambda e: XEtaProductAlgebra(None, QQ)(XEtaProducts()(e))

        self.assertEqual( repr(XEtaProductAlgebra(None, QQ).zero()),
                          "" )
        self.assertIn( repr(eps((0,2,1,0)) + eps((1,3,1,0))**2),
                       ["1*Lambda_{2,1,0}(z) + 1*eta(3z)^2",
                        "1*eta(3z)^2 + 1*Lambda_{2,1,0}(z)"] )

    def test___hash__(self) :
        eps = lambda e: XEtaProductAlgebra(QQ)(XEtaProducts()(e))

        self.assertEqual( hash(eps((0,2,1,0)) + eps((1,3,1,0))**2),
                          2823831037481720944 )
        self.assertEqual( hash(eps((0,2,1,5)) + eps((1,3,1,3))**4),
                          -6992706302413729911 )

class TestXEtaProductAlgebra(unittest.TestCase) :

    def setUp(self) :
        pass

    def test_class(self) :
        self.assertEqual( XEtaProductAlgebra().category(),
                          CommutativeAlgebras(QQ) )

        self.assertIsNone( TestSuite(XEtaProductAlgebra).run() )

    def test___init__(self) :
        self.assertEqual( XEtaProductAlgebra().level(),
                          1 )
        self.assertEqual( XEtaProductAlgebra().base_ring(),
                          QQ )
        
        self.assertEqual( XEtaProductAlgebra(7).level(),
                          7 )
        self.assertEqual( XEtaProductAlgebra(7, CyclotomicField(9)).base_ring(),
                          CyclotomicField(9) )
        self.assertEqual( XEtaProductAlgebra(base_ring = CyclotomicField(9)).base_ring(),
                          CyclotomicField(9) )

    def test_level(self) :
        self.assertEqual( XEtaProductAlgebra(8).level(),
                          8 )

    def test__an_element(self) :
        self.assertIn( XEtaProductAlgebra(7, CyclotomicField(9)).an_element(),
                       XEtaProductAlgebra(7, CyclotomicField(9)) )

    def test_construction(self) :
        h = lambda (f,r) : f(r)
        self.assertEqual( h(XEtaProductAlgebra(7, CyclotomicField(9)).construction()),
                          XEtaProductAlgebra(7, CyclotomicField(9)) )

    def test__has_coerce_map_from_(self) :
        self.assertTrue( XEtaProductAlgebra(None, QQ).has_coerce_map_from(XEtaProductAlgebra(2, ZZ)) )
        self.assertTrue( XEtaProductAlgebra(4, QQ).has_coerce_map_from(XEtaProductAlgebra(2, ZZ)) )
        self.assertFalse( XEtaProductAlgebra(3, QQ).has_coerce_map_from(XEtaProductAlgebra(2, ZZ)) )
        self.assertFalse( XEtaProductAlgebra(3, QQ).has_coerce_map_from(XEtaProductAlgebra(3, CyclotomicField(4))) )

    def test__element_constructor_(self) :
        eps = XEtaProducts()
        epsa = XEtaProductAlgebra(7, CyclotomicField(9))

        self.assertIs( epsa(eps((0,2,1,0))).parent(),
                       epsa )
        self.assertIs( epsa({ eps((0,2,1,0)): 1, eps((0,3,2,1)): 3}).parent(),
                       epsa )
        self.assertIs( epsa(3).parent(),
                       epsa )

        self.assertEqual( epsa({ eps((0,2,1,0)): 1, eps((0,3,2,1)): 3 }).parent(),
                          epsa({ eps((0,2,1,0)): CyclotomicField(9)(1),
                                 eps((0,3,2,1)): 3 }).parent() )
        self.assertEqual( epsa(3),
                          epsa(CyclotomicField(9)(3)) )

    def test__repr_(self) :
        self.assertEqual( repr(XEtaProductAlgebra(7, CyclotomicField(9))),
                          "Algebra of extended eta products of level 7 over Cyclotomic Field of order 9 and degree 6" )

class TestXEtaProductAlgebra_functor(unittest.TestCase) :

    def setUp(self) :
        pass

    def test___init__(self) :
        self.assertEqual( XEtaProductAlgebra_functor(7).domain(),
                          CommutativeRings() )
        self.assertEqual( XEtaProductAlgebra_functor(7).codomain(),
                          CommutativeRings() )

        self.assertEqual( XEtaProductAlgebra_functor(7)._level,
                          7 )

    def test_merge(self) :
        self.assertEqual( XEtaProductAlgebra_functor(7).merge(XEtaProductAlgebra_functor(7)),
                          XEtaProductAlgebra_functor(7) )
 
        self.assertEqual( XEtaProductAlgebra_functor(7).merge(XEtaProductAlgebra_functor(3)),
                          XEtaProductAlgebra_functor(21) )
        self.assertEqual( XEtaProductAlgebra_functor(3).merge(XEtaProductAlgebra_functor(7)),
                          XEtaProductAlgebra_functor(21) )
        self.assertEqual( XEtaProductAlgebra_functor(3).merge(XEtaProductAlgebra_functor(None)),
                          XEtaProductAlgebra_functor(None) )

    def test__apply_functor(self) :
        self.assertEqual( XEtaProductAlgebra_functor(7)(CyclotomicField(9)),
                          XEtaProductAlgebra(7, CyclotomicField(9)) )

class TestHelperFunctions(unittest.TestCase) :

    def setUp(self) :
        pass

    def test__divide_q_power(self) :
        FEQ = PowerSeriesRing(QQ, 'q')
        q = FEQ.gen()

        self.assertTrue( _divide_q_power(q**2, 1).parent() is FEQ )

        self.assertEqual( _divide_q_power(q**2 + 4 * q**3 + 7 * q**4, 2),
                          q + 7 * q**2 )

        R = PowerSeriesRing(QQ, 'x')
        x = R.gen()
        self.assertEqual( _divide_q_power(x**2 + 4 * x**3 + 7 * x**4 + O(x**8), 2),
                          x + 7 * x**2 + O(x**4) )

    def test__eta(self) :
        FEQ = PowerSeriesRing(QQ, 'q')
        pari.set_series_precision(20)

        self.assertEqual( _eta(20),
                          FEQ(pari('eta(q)')) )

        self.assertEqual( _eta(20).parent(),
                          FEQ )

    def test__eta_d(self) :
        FEQ = PowerSeriesRing(QQ, 'q')
        q = FEQ.gen()

        pari.set_series_precision(20)
        self.assertEqual( _divide_q_power(
                           FEQ(dict( (e, 24 * v / Integer(e))
                                     for (e, v) in _eta_d(20).polynomial().subs(q = q**24).shift(1) \
                                         .dict().iteritems() )).shift(-1),
                           24),
                          FEQ(pari('eta(q)')) )
                          
        self.assertEqual( _eta_d(20).parent(),
                          PowerSeriesRing(QQ, 'q') )

    def test__zeta_denominator_xep(self) :
        M = Integer(randint(1, 200))
        N = Integer(randint(1, 200))

        self.assertEqual( _zeta_denominator_xep([(1,M,N,0)]),
                          1 )
        self.assertEqual( _zeta_denominator_xep([(-1,M,N,0)]),
                          1 )
        self.assertEqual( _zeta_denominator_xep([(0,M,N,0)]),
                          1 )

        self.assertEqual( _zeta_denominator_xep([(1,M,1,1)]),
                          24 )
        self.assertEqual( _zeta_denominator_xep([(-1,M,1,1)]),
                          24 )
        self.assertEqual( _zeta_denominator_xep([(0,M,N,1)]),
                          N )

        self.assertEqual( _zeta_denominator_xep([(1,M,1,1), (-1,M,1,1)]),
                          1 )
        self.assertEqual( _zeta_denominator_xep(24 * [(1,M,1,1)]),
                          1 )
        self.assertEqual( _zeta_denominator_xep((12 * N) * [(1,M,N,0)]),
                          1 )
        self.assertEqual( _zeta_denominator_xep((12 * N) * [(1,M,N,2)]),
                          N )

        self.assertEqual( _zeta_denominator_xep([(1,M,N,1)]),
                          24 * N )
        self.assertEqual( _zeta_denominator_xep([(-1,M,N,1)]),
                          24 * N )
        self.assertEqual( _zeta_denominator_xep([(0,M,N,1), (1,M,2*N,3)]),
                          16 * N )

    def test__eval_xep__complex(self) :
        self.assertEqual( _eval_xep__complex([(0,1,1,0)], 20, CyclotomicField(10))[0],
                          0 )

        self.assertEqual( _eval_xep([(0,5,1,0)], 20)[0],
                          _eval_xep__complex([(0,5,1,0)], 20, CyclotomicField(10))[0] )
        self.assertEqual( _eval_xep([(0,1,5,0)], 20)[0],
                          _eval_xep__complex([(0,1,5,0)], 20, CyclotomicField(10))[0] )

        e = ((-1, 1, 1, 0), (-1, 1, 1, 0), (-1, 6, 1, 0), (-1, 6, 2, 1), (-1, 6, 2, 1), (-1, 6, 3, 1), (1, 2, 1, 0), (1, 2, 1, 0), (1, 2, 2, 1), (1, 2, 2, 1), (1, 2, 2, 1), (1, 2, 2, 1), (1, 3, 1, 0), (1, 3, 1, 0), (1, 6, 6, 1), (1, 6, 6, 1))
        self.assertEqual( _eval_xep(e, 20, CyclotomicField(24)),
                          _eval_xep__complex(e, 20, CyclotomicField(24)) )

    def test__eval_xep(self) :
        self.assertEqual( _eval_xep([(0,1,1,0)], 20)[0],
                          0 )
        self.assertRaises( ValueError,
                           _eval_xep, [(0,1,20,1)], 20 )

        ## the inverse of an eta product and the eta product itself must cancel
        self.assertEqual( _eval_xep(2*[(-1,1,1,0), (1,2,1,0), (1,4,1,0),
                                       (1,5,1,0), (1,10,1,0), (-1,20,1,0)], 27)[0]**-1,
                          _eval_xep(2*[(1,1,1,0), (-1,2,1,0), (-1,4,1,0),
                                       (-1,5,1,0), (-1,10,1,0), (1,20,1,0)], 27)[0] )

        ## some cases where the Lambda_N are sufficient to span the
        ## space of Eisenstein series
        self.assertEqual( _eval_xep([(0,5,1,0)], 20),
                          (Integer(5)/6 * EisensteinForms(5, 2).gen(0).qexp(20), 1) )
        self.assertEqual( _eval_xep([(0,7,1,0)], 20),
                          (Integer(7)/4 * EisensteinForms(7, 2).gen(0).qexp(20), 1) )
        lambda_span = span(vector(_eval_xep([(0,N,1,0)], 20)[0].padded_list(20)) for N in [2, 5, 10])
        eis_span = span(vector(b.qexp(20).padded_list(20)) for  b in EisensteinForms(10, 2).basis())
        self.assertEqual( lambda_span,
                          eis_span )
        
        ## some cases where the Eisenstein series is unique, so that the expansions at
        ## another cusp can be computed directly
        self.assertEqual( _eval_xep([(0,1,5,0)], 4),
                          (Integer(-1)/6 * EisensteinForms(5, 2).gen(0).qexp(20), 5) )
        self.assertEqual( _eval_xep([(0,1,7,0)], 4),
                          (Integer(-1)/4 * EisensteinForms(7, 2).gen(0).qexp(28), 7) )

        ## spaces with unique cusp forms
        self.assertEqual( _eval_xep(2*[(1,1,1,0)] + 2*[(1,11,1,0)], 20),
                          (CuspForms(11, 2).gen(0).qexp(20), 1) )
        self.assertEqual( _eval_xep([(1,1,1,0), (1,2,1,0), (1,7,1,0), (1,14,1,0)], 20),
                          (CuspForms(14, 2).gen(0).qexp(20), 1) )

        ## we use the above spaces to check other cusp expansions
        self.assertEqual( _eval_xep(2*[(1,1,1,0)] + 2*[(1,1,11,0)], 4),
                          (CuspForms(11, 2).gen(0).qexp(44), 11) )
        self.assertEqual( _eval_xep([(1,1,1,0), (1,1,2,0), (1,1,7,0), (1,1,14,0)], 3),
                          (CuspForms(14, 2).gen(0).qexp(42), 14) )

        ## some eta products and their transforms that we check for modularity
        mf_span = span( vector(b.qexp(27).padded_list(27)) for b in ModularForms(9, 2).basis() )
        self.assertIn( vector(_eval_xep(2*[(-1,3,1,0)] + 6*[(1,9,1,0)], 27)[0].padded_list(27)),
                       mf_span )
        self.assertIn( vector(_eval_xep(2*[(-1,3,1,0)] + 6*[(1,9,1,0)], 27)[0].padded_list(27)),
                       mf_span )
        self.assertIn( vector(_divide_q_power(
                                _eval_xep(2*[(-1,3,1,0)] + 6*[(1,3,3,1)], 27, CyclotomicField(3 * 24))[0], 3) \
                              .padded_list(27)),
                       mf_span.change_ring(CyclotomicField(3 * 24)) )

        ## a product of Lambdas that we check for modularity
        mf_span = span( vector(b.qexp(20).padded_list(20)) for b in ModularForms(7, 6).basis() )
        self.assertIn( vector(_eval_xep(3*[(0,7,1,0)], 20)[0].padded_list(20)),
                       mf_span )
        mf_span = span( vector(b.qexp(20).padded_list(20)) for b in ModularForms(14, 4).basis() )
        self.assertIn( vector(_eval_xep([(0,2,1,0), (0,7,1,0)], 20)[0].padded_list(20)),
                       mf_span )

    ## We don't test _eval_xep__with_zeta, which should be sufficiently well tested
    ## by the methods which call it.

    def test__eval_projection_xep(self) :
        self.assertEqual( _eval_projection_xep([(0,1,1,0)], 20),
                          0 )
        self.assertEqual( _eval_projection_xep([(1,3,1,0)], 20),
                          0 )
        self.assertRaises( ValueError,
                           _eval_projection_xep, [(0,1,20,1)], 20 )

        ## check equality with the usual evaluation in case that lN = 1
        self.assertEqual( _eval_xep([(0,7,1,0)], 20)[0],
                          _eval_projection_xep([(0,7,1,0)], 20) )
        self.assertEqual( _eval_xep(2*[(-1,1,1,0), (1,2,1,0), (1,4,1,0),
                                       (1,5,1,0), (1,10,1,0), (-1,20,1,0)], 27)[0],
                          _eval_projection_xep(2*[(-1,1,1,0), (1,2,1,0), (1,4,1,0),
                                                  (1,5,1,0), (1,10,1,0), (-1,20,1,0)], 27) )

        ## some spaces where the old spaces are small, so that the projections can be easily
        ## identified
        self.assertEqual( -3 * _eval_projection_xep([(0,1,9,0)], 20),
                          _divide_q_power(ModularForms(9,2).echelon_basis()[0].qexp(60), 3) )
        self.assertEqual( -8 * _eval_projection_xep([(0,1,4,1)], 20, CyclotomicField(4 * 24)),
                          _divide_q_power(ModularForms(4,2).echelon_basis()[0].qexp(40), 2) )


    def test__cusp_matrix(self) :
        self.assertEqual( _cusp_matrix(Cusp(infinity)),
                          (1,0,0,1) )
        for _ in range(10) :
            cusp = QQ.random_element()
            (a, b, c, d) = _cusp_matrix(cusp)
            
            self.assertTrue( a * d - b * c, 1 )
            self.assertTrue( QQ((a,c)) == cusp )
    
    def test__cusp_expansion(self) :
        M = Integer(randint(1,200))
        m = SL2Z.random_element()
        minv = m**-1

        self.assertRaises( NotImplementedError,
                           _cusp_expansion, (0, 1, 3, 0), tuple(m) )
        self.assertRaises( NotImplementedError,
                           _cusp_expansion, (0, 3, 1, 2), tuple(m) )


        self.assertEqual( _cusp_expansion( (0, M, 1, 0), (1,0,0,1) )[0],
                          (0, M, 1, 0) )
        self.assertEqual( _cusp_expansion( (0, M, 1, 0), (0,-1,1,0) )[0],
                          (0, 1, M, 0) )

        self.assertEqual( _cusp_expansion((0,M,1,0), (1,1,0,1))[1],
                          0 )
        self.assertEqual( _cusp_expansion((0,M,1,0), (1,-1,0,1))[1],
                          0 )
        self.assertEqual( _cusp_expansion((1,M,1,0), (1,1,0,1))[1],
                          0 )
        self.assertEqual( _cusp_expansion((1,1,1,0), (0,-1,1,0))[1],
                          -Integer(1)/8 )
        self.assertEqual( _cusp_expansion((-1,1,1,0), (0,-1,1,0))[1],
                          Integer(1)/8 )

        # (M, N, h) : Scaling parameters for an eta function `\eta((Mz + h) / N)
        # c : A cusp
        def test_numerical((M, N, h), c) :
            (a_g, b_g, c_g, d_g) = _cusp_matrix(c)
            gamma = lambda tau: (a_g * tau + b_g) / (c_g * tau + d_g)
            def j(tau) :
                res = CC(c_g * tau + d_g).sqrt()
                if res.real_part() < 0 :
                    return -res
                elif res.real_part() == 0 and res.imag_part() > 0 :
                    return -res
                else :
                    return res

            ((_, cM, cN, ch), eta_mult) = _cusp_expansion((1, M, N, h), (a_g, b_g, c_g, d_g))

            eta_qexp = _eta(5000)

            tau = SR('I') + 1/3
            e_tau = (M * gamma(tau) + h) / Integer(N)
            ce_tau = (cM * tau + ch) / Integer(cN)
            e_eval = CC(   j(tau)**-1
                         * CC(N).sqrt()**-1
                         * CC(exp(2 * pi * I * e_tau / 24))
                         * eta_qexp.polynomial().subs(q = CC(exp(2 * pi * I * e_tau))) )
            ce_eval = CC(   CC(cN).sqrt()**-1
                          * exp(2 * pi * I * eta_mult)
                          * CC(exp(2 * pi * I * ce_tau / 24))
                          * eta_qexp.polynomial().subs(q = CC(exp(2 * pi * I * ce_tau))) )
 
            self.assertAlmostEqual( e_eval,
                                    ce_eval )

        test_numerical((1,1,0), Cusp(0))
        test_numerical((1,1,0), Cusp(1))
        test_numerical((1,1,0), Cusp(2))
        test_numerical((1,1,0), Cusp(7))
        test_numerical((1,1,0), Cusp(Integer(1)/2))
        test_numerical((1,1,0), Cusp(Integer(1)/3))
        test_numerical((1,1,0), Cusp(Integer(1)/4))
        test_numerical((1,1,0), Cusp(Integer(1)/5))
        test_numerical((1,1,0), Cusp(Integer(2)/3))
        test_numerical((1,1,0), Cusp(Integer(7)/15))
        test_numerical((2,1,0), Cusp(0))
        test_numerical((2,1,0), Cusp(Integer(1)/2))
        test_numerical((4,1,0), Cusp(0))
        test_numerical((4,1,0), Cusp(Integer(1)/2))
        test_numerical((3,1,0), Cusp(0))
        test_numerical((3,1,0), Cusp(Integer(1)/3))
        test_numerical((9,1,0), Cusp(0))
        test_numerical((9,1,0), Cusp(Integer(1)/3))

        test_numerical((23,1,0), Cusp(0))
        test_numerical((46,1,0), Cusp(0))
        
    def test__eta_multiplier(self) :
        for _ in range(10) :
            m = SL2Z.random_element()
            minv = m**-1
            e = _eta_multiplier(*tuple(m))

            self.assertTrue( 2 * (_eta_multiplier(*tuple(minv)) + e) in ZZ )
            self.assertTrue( 24 * e in ZZ )

        def test_numerical(M, (a_g, b_g, c_g, d_g)) :
            CC = ComplexField(200)
            gamma = lambda tau: (a_g * tau + b_g) / (c_g * tau + d_g)
            def j(tau) :
                res = CC(c_g * tau + d_g).sqrt()
                if res.real_part() < 0 :
                    return -res
                elif res.real_part() == 0 and res.imag_part() > 0 :
                    return -res
                else :
                    return res
            eta_mult = _eta_multiplier(a_g, M * b_g, c_g / M, d_g)

            eta_qexp = _eta(24 * 2000).polynomial()

            tau = SR('I') + 1/3

            e_eval = CC(   j(tau)**-1
                         * CC(exp(2 * pi * I * M * gamma(tau) / 24))
                         * eta_qexp.subs(q = CC(exp(2 * pi * I * M * gamma(tau) ))) )
            ce_eval = CC(   exp(2 * pi * I * eta_mult)
                          * CC(exp(2 * pi * I * M * tau / 24))
                          * eta_qexp.subs(q = CC(exp(2 * pi * I * M * tau))) )

            self.assertAlmostEqual( e_eval,
                                    ce_eval )

        from sage.modular.all import FareySymbol
        for g in FareySymbol(Gamma0(11)).generators() :
            test_numerical(1, list(g))
            test_numerical(11, list(g))
        for g in FareySymbol(Gamma0(21)).generators() :
            test_numerical(1, list(g))
            test_numerical(3, list(g))
            test_numerical(7, list(g))
            test_numerical(21, list(g))

    def test__square_normalization_xep(self) :
        M = Integer(randint(1, 200))
        N = Integer(randint(1, 200))
        h = Integer(randint(-200, 200))

        self.assertEqual( _square_normalization_xep([(1,M,N,h)]),
                          1/N )
        self.assertEqual( _square_normalization_xep(5 * [(1,M,N,h)]),
                          1/N**5 )
        self.assertEqual( _square_normalization_xep(5 * [(1,M,N,h)] + [(-1, 1, N, 1)]),
                          1/N**4 )

        self.assertEqual( _square_normalization_xep([(0,M,N,h)]),
                          1 )

if __name__ == '__main__':
    #suite = unittest.TestSuite([TestHelperFunctions("test_modularity")])
    #unittest.TextTestRunner(verbosity=2).run(suite)

    unittest.main(verbosity = 2)
